#pragma once

#include <amdis/GridFunctions.hpp>
#include <amdis/gridfunctions/DiscreteFunction.hpp>

namespace AMDiS
{
  /// A mutable view on the subspace of a DOFVector, \relates DiscreteFunction
  template <class GB, class VT, class TP>
  class DOFVectorView
      : public DiscreteFunction<GB, VT, TP>
  {
    using Self = DOFVectorView;
    using Super = DiscreteFunction<GB, VT, TP>;

    using GlobalBasis = GB;
    using TreePath = TP;

  public:
    /// Constructor forwards to the treePath constructor, with empty TreePath
    DOFVectorView(DOFVector<GB,VT>& dofVector)
      : DOFVectorView{dofVector, Dune::TypeTree::hybridTreePath()}
    {}

    /// Constructor. Stores a pointer to the mutable `dofvector`.
    DOFVectorView(DOFVector<GB,VT>& dofVector, TP const& treePath)
      : Super(dofVector, treePath)
      , mutableDofVector_(&dofVector)
    {}

  public:
    /// \brief Interpolation of GridFunction to DOFVector, assuming that there is no
    /// reference to this DOFVector in the expression.
    /**
     * **Example:**
     * ```
     * auto v = makeDOFVectorView(prob.solutionVector(),0);
     * v.interpolate_noalias([](auto const& x) { return x[0]; });
     * ```
     **/
    template <class Expr>
    void interpolate_noalias(Expr&& expr)
    {
      auto const& basis = this->basis();
      auto const& treePath = this->treePath();

      auto&& gridFct = makeGridFunction(FWD(expr), basis.gridView());

      Dune::Functions::interpolate(basis, treePath, coefficients(), FWD(gridFct));
    }

    /// \brief Interpolation of GridFunction to DOFVector
    /**
     * **Example:**
     * ```
     * auto v = makeDOFVectorView(prob.solutionVector(),0);
     * v.interpolate(v + [](auto const& x) { return x[0]; });
     * ```
     * Allows to have a reference to the DOFVector in the expression, e.g. as
     * \ref DiscreteFunction or \ref gradientAtQP() of a DiscreteFunction.
     **/
    template <class Expr>
    void interpolate(Expr&& expr)
    {
      // create temporary copy of data
      DOFVector<GB,VT> tmp(coefficients());
      Self tmpView{tmp, this->treePath()};
      tmpView.interpolate_noalias(FWD(expr));

      // move data from temporary vector into stored DOFVector
      coefficients().vector() = std::move(tmp.vector());
    }

    /// \brief Interpolation of GridFunction to DOFVector, alias to \ref interpolate()
    template <class Expr>
    DOFVectorView& operator<<(Expr&& expr)
    {
      interpolate(expr);
      return *this;
    }

    /// \brief interpolate `(*this) + expr` to DOFVector
    template <class Expr>
    DOFVectorView& operator+=(Expr&& expr)
    {
      interpolate((*this) + expr);
      return *this;
    }

    /// \brief interpolate `(*this) - expr` to DOFVector
    template <class Expr>
    DOFVectorView& operator-=(Expr&& expr)
    {
      interpolate((*this) - expr);
      return *this;
    }

    /// Return the mutable DOFVector
    DOFVector<GB,VT>& coefficients() { return *mutableDofVector_; }

    /// Return the const DOFVector
    using Super::coefficients;

  protected:
    DOFVector<GB,VT>* mutableDofVector_;
  };


#if DUNE_HAVE_CXX_CLASS_TEMPLATE_ARGUMENT_DEDUCTION
  // Deduction guide for DOFVectorView class
  template <class GlobalBasis, class ValueType>
  DOFVectorView(DOFVector<GlobalBasis, ValueType>& dofVector)
    -> DOFVectorView<GlobalBasis, ValueType, Dune::TypeTree::HybridTreePath<>>;
#endif

  /// A Generator for a mutable \ref DOFVectorView
  template <class GlobalBasis, class ValueType, class PreTreePath>
  auto makeDOFVectorView(DOFVector<GlobalBasis, ValueType>& dofVector, PreTreePath const& preTreePath)
  {
    auto treePath = makeTreePath(preTreePath);
    return DOFVectorView<GlobalBasis, ValueType, decltype(treePath)>{dofVector, treePath};
  }

  /// A Generator for a mutable \ref DOFVectorView
  template <class GlobalBasis, class ValueType>
  auto makeDOFVectorView(DOFVector<GlobalBasis, ValueType>& dofVector)
  {
    auto treePath = Dune::TypeTree::hybridTreePath();
    return DOFVectorView<GlobalBasis, ValueType, Dune::TypeTree::HybridTreePath<>>{dofVector, treePath};
  }

} // end namespace AMDiS
