#pragma once

#include <dune/istl/preconditioner.hh>

#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>
#include <amdis/linearalgebra/istl/Fwd.hpp>
#include <amdis/linearalgebra/istl/ISTL_Preconditioner.hpp>

namespace AMDiS
{
  template <class ISTLSolver, class Matrix, class VectorX, class VectorB>
  class ISTLRunner
      : public RunnerInterface<Matrix, VectorX, VectorB>
  {
    using LinOperator = Dune::MatrixAdapter<Matrix, VectorX, VectorB>;
    using BaseSolver  = Dune::InverseOperator<VectorX, VectorB>;
    using BasePrecon  = Dune::Preconditioner<VectorX, VectorB>;
    using ISTLPrecon  = ISTLPreconInterface<Matrix, VectorX, VectorB>;

  public:
    ISTLRunner(std::string const& prefix)
      : solverCreator_(prefix)
    {
      initPrecon(prefix);
    }


    /// Implementes \ref RunnerInterface::init()
    void init(Matrix const& A) override
    {
      precon_ = preconCreator_->create(A);
      linOperator_ = std::make_shared<LinOperator>(A);
      solver_ = solverCreator_.create(*linOperator_, *precon_);
    }

    /// Implementes \ref RunnerInterface::exit()
    void exit() override
    {
      solver_.reset();
      linOperator_.reset();
      precon_.reset();
    }

    /// Implementes \ref RunnerInterface::solve()
    int solve(Matrix const& A, VectorX& x, VectorB const& b,
              SolverInfo& solverInfo) override
    {
      // storing some statistics
      Dune::InverseOperatorResult statistics;

      // solve the linear system
      VectorB _b = b;
      solver_->apply(x, _b, statistics);

      solverInfo.setRelResidual(statistics.reduction);

      return statistics.converged ? 0 : 1;
    }

  protected:
    void initPrecon(std::string const& prefix)
    {
      // get creator string for preconditioner
      std::string preconName = "no";
      std::string initFileStr = "";
      for (std::string postfix : {"left precon", "right precon", "precon->name"}) {
        Parameters::get(prefix + "->" + postfix, preconName);
        if (!preconName.empty() && preconName != "no") {
          initFileStr = prefix + "->" + postfix;
          break;
        }
      }

      if (preconName.empty() || preconName == "no")
        preconName = "default";

      auto* creator = named(CreatorMap<ISTLPrecon>::getCreator(preconName, initFileStr));
      preconCreator_ = creator->createWithString(initFileStr);
      assert(preconCreator_);
    }

  private:
    ISTLSolverCreator<ISTLSolver> solverCreator_;
    std::shared_ptr<ISTLPrecon> preconCreator_;

    std::shared_ptr<BasePrecon>   precon_;
    std::shared_ptr<LinOperator>  linOperator_;
    std::shared_ptr<BaseSolver>   solver_;
  };

} // end namespace AMDiS
