#pragma once

#include <type_traits>

#include <amdis/localoperators/FirstOrderTestPartialTrial.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct partialtest
    {
      std::size_t comp;
    };
  }


  /// first-order operator \f$ \langle\partial_i\psi, c\rangle \f$

  template <class LC, class GridFct>
  class GridFunctionOperator<tag::partialtest, LC, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::partialtest, LC, GridFct>, LC, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LC, GridFct>;

    static_assert( Size_v<typename GridFct::Range> == 1, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::partialtest tag, GridFct const& expr)
      : Super(expr, 1)
      , comp_(tag.comp)
    {}

    template <class CG, class Node, class Vec>
    void getElementVector(CG const& contextGeo, Node const& node, Vec& elementVector)
    {
      static_assert(Node::isLeaf, "Operator can be applied to Leaf-Nodes only.");

      auto const& quad = this->getQuadratureRule(contextGeo.type(), node);
      auto const& localFE = node.finiteElement();
      std::size_t feSize = localFE.size();

      using RangeFieldType = typename NodeQuadCache<Node>::Traits::RangeFieldType;
      NodeQuadCache<Node> cache(localFE.localBasis());

      auto const& shapeGradientsCache = cache.evaluateJacobianAtQP(contextGeo, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = contextGeo.geometry().jacobianInverseTransposed(local);
        assert(jacobian.M() == CG::dim);

        // The multiplicative factor in the integral transformation formula
        const auto dx = contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();
        const auto c = Super::coefficient(local);

        // The gradients of the shape functions on the reference element
        auto const& shapeGradients = shapeGradientsCache[iq];

        // Compute the shape function gradients on the real element
        thread_local std::vector<RangeFieldType> partial;
        partial.resize(shapeGradients.size());
        for (std::size_t i = 0; i < partial.size(); ++i) {
          partial[i] = jacobian[comp_][0] * shapeGradients[i][0][0];
          for (std::size_t j = 1; j < jacobian.M(); ++j)
            partial[i] += jacobian[comp_][j] * shapeGradients[i][0][j];
        }

        for (std::size_t j = 0; j < feSize; ++j) {
          const auto local_j = node.localIndex(j);
          elementVector[local_j] += dx * (c * partial[j]);
        }
      }
    }

  private:
    std::size_t comp_;
  };

  /** @} **/

} // end namespace AMDiS
